from openai import OpenAI
from dotenv import load_dotenv
import os
import rich


class BugFixer:
    def __init__(self, function, function_name, inputs, expected, returns, api_key):
        inputs = ", ".join([f"{key}={value}" for key, value in inputs.items()])
        self.prompt = f"function:\n{function}\nthe function name is: {function_name}\nthe expected value is {expected} when the input is: {inputs}\nthe function actually returns{' the error' if returns[1] else ''}: {returns[0]}"
        load_dotenv()
        self.client = OpenAI(api_key=os.getenv(api_key))
        self.instructions = "fix the following code and response just and exclusively with the fixed code using the provided data, no explanations, retrieve the whole function"
    def get_fixed_function(self):        
        completion = self.client.chat.completions.create(
        model="gpt-3.5-turbo-1106",
        messages=[
            {"role": "system", "content": self.instructions},
            {"role": "user", "content": self.prompt},
        ]
        )
        return "\n".join(completion.choices[0].message.content.replace('\\n', '\n').replace('```','').splitlines()[1:])




def fix_bug(function_name, path, data, inputs, function, main_response, args, get_module):
    if function[function_name].get('use-ai') or function[function_name].get('auto-fix'):
        if function[function_name]["expected"].get("comparator") != "Equals":
            rich.print(f"[bold red]Test for {function_name}: Cannot use ai with assertion other than Equals[/bold red]")
            return
        if not data.get("api-key"):
            rich.print("[bold red]Cannot use ai without OpenAI api-key[/bold red]")
            return
        
        func_code = get_module('utils').extract_function_code(function_name, path)
        if not func_code:
            rich.print(f"[bold red]Test for {function_name}: Cannot find function code[/bold red]")
            return
        bug_fixer = get_module('ai').BugFixer(func_code, function_name, inputs, function[function_name]["expected"]["value"], [main_response if isinstance(main_response, str) else main_response[0], isinstance(main_response, str)], data["api-key"])
        if isinstance(main_response, str):
            bug = main_response
        else:
            bug = f"{main_response[0]} is not {function[function_name]['expected']['value']}"
        
        if function[function_name].get('auto-fix'):
            rich.print(f"[bold yellow]Auto fixing bug: '{bug}' in '{function_name}' using AI[/bold yellow]")
        else:
            fix_bug = input(f"Do you want to fix the bug: '{bug}' in '{function_name}' using AI? (Y/n) ")
            if fix_bug.lower() != "y":
                return
        fixed_bug = bug_fixer.get_fixed_function()
        with open(f"fixed_bug_{function_name}.tests.py", "w") as f:
            f.write(fixed_bug)
        
        print_return = function[function_name].get("print-return") or args.verbose or function[function_name].get("verbose")
        
        
        
        args_ditctionary = {
            "file": f"fixed_bug_{function_name}.tests.py",
            "function": function_name,
            "params": inputs,
            "assertion": function[function_name]["expected"]["comparator"],
            "assert_to": function[function_name]["expected"]["value"],
            "print_return": print_return
            
        }
        if function[function_name].get('instance'):
            args_ditctionary.update({"instance_params":function[function_name]["instance"]})
        
        args_for_main = get_module('utils').Args(args_ditctionary)
        
        test_fixed_bug = get_module('tester').test(args_for_main)
        if isinstance(test_fixed_bug, str):
            rich.print("[bold red]Bug fixed failed[/bold red]")
        rich.print("[bold green]Bug fixed successfully[/bold green]")
        if function[function_name].get('auto-fix'):
            rich.print("[bold yellow]Auto replacing the original code with the fixed code[/bold yellow]")
        else:
            replace = input("Do you want to replace the original code with the fixed code? (Y/n) ")
            if replace.lower() != "y":
                return
        all_code = open(path,'r').read()
        all_code = all_code.replace(func_code, fixed_bug)
        with open(path, 'w') as f:
            f.write(all_code)
        rich.print("[bold green]Code replaced successfully[/bold green]")
        os.remove(f"fixed_bug_{function_name}.tests.py")
        return test_fixed_bug
    return main_response


import re


def extract_function_code(function_name, source_file):
    # Build the regex pattern to find the function
    if '.' in function_name:
        function_name = function_name.split('.')[1]
    function_pattern = re.compile(fr'(\bdef {function_name}\(.*?\):)(?:\n(.*?)(?=\n\s*def|\n\s*class|\n*$))?', re.DOTALL)
    
    with open(source_file, 'r') as file:
        content = file.read()

    # Search for the function in the file content
    matches = function_pattern.search(content)
    if matches:
        # The function was found, return the code
        function_declaration = matches.group(1)
        function_code = matches.group(2)
        function_code = "\n".join([function_declaration, function_code])
        
        lines = function_code.split('\n')
        line2 = lines[1]
        indentation = re.split(r'\w+', line2)[0]
        indentation = indentation.count(' ') + indentation.count('\t')
        
        function_code = f"{lines[0]}\n{lines[1]}"
        for line in lines[2:]:
            current_indentation = re.split(r'\w+', line)[0]
            current_indentation = current_indentation.count(' ') + current_indentation.count('\t')
            if current_indentation < indentation:
                break
            else:
                function_code += f"\n{line}"
        return function_code
    else:
        # The function was not found
        return None


class Args:
    def __init__(self, args):
        for key, value in args.items():
            setattr(self, key, value)

class Kwargs:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

def params_to_dict(params: list) -> dict:
    dict_params = {}
    for i in range(0,len(params),2):
        param = {params[i]:params[i+1]}
        dict_params.update(param)

    return dict_params
def eval_params_values(params: dict,functions: dict) -> dict:
    for key,value in params.items():
        if functions[key] in [int,float]:
            params[key] = functions[key](value)
        elif functions[key] != str:
            params[key] = eval(value)
        else:
            params[key] = f'{value}'
    return params

class OutputArgs:
    def __init__(self, title: str, groups: list = []):
        self.title = title
        self.groups = groups
    def add_group(self, groups: dict):
        self.groups.append(groups)
    def add_test(self, test: dict, group_name: str):
        for group in self.groups:
            if group["name"] == group_name:
                group["tests"].append(test)


import rich
from . import info

def output_classic(args):
    rich.print(f"[bold cyan]{args.title}[/bold cyan]")
    for group in args.groups:
        rich.print(f"[bold blue]{' '*4}- {group['name']}[/bold blue]")
        for test in group["tests"]:
            rich.print(f"[bold magenta]{' '*8}- {test['name']}[/bold magenta]")
            if test.get('ignore'):
                rich.print(f"[bold red]{' '*12}-Test: ignored[/bold red]")
                continue
            if not test['verbose']:
                if test['status']['passed']:
                    rich.print(f"[bold green]{' '*12}-Test: passed[/bold green]")
                else:
                    returned = test['status']['reason']
                    returned = f"'{returned}'" if isinstance(returned, str) else returned
                    rich.print(f"[bold red]{' '*12}-Test: failed: {returned}[/bold red]")
                continue
            if test['params'] != {}:
                rich.print(f"[bold yellow]{' '*12}- inputs:[/bold yellow]")
                for key, value in test['params'].items():
                    value = f"'{value}'" if isinstance(value, str) else value
                    rich.print(f"[bold yellow]{' '*14}- {key}: {value}[/bold yellow]")
            if test.get('instance_params'):
                rich.print(f"[bold yellow]{' '*12}- instance:[/bold yellow]")
                for key, value in test['instance_params'].items():
                    value = f"'{value}'" if isinstance(value, str) else value
                    rich.print(f"[bold yellow]{' '*14}- {key}: {value}[/bold yellow]")
            assert_to = test.get('assert_to', None)
            if not assert_to:
                assert_to_dot = test.get('assert_to_dot', None)
                if not assert_to_dot:
                    rich.print(f"[bold red]{' '*12}-Cannot understand expected.value[/bold red]")
                    rich.print(f"[bold red]{' '*12}-Test: failed[/bold red]")
                    continue
                assert_to_dot = [assert_to_dot[0], ".".join([str(_assert) for _assert in assert_to_dot[1]])]
                assert_to_dot[1] = assert_to_dot[1].replace(';',':')
                assert_to = f"'{assert_to_dot[0]}'; in value.{assert_to_dot[1]}" if isinstance(assert_to_dot[0], str) else f"{assert_to_dot[0]}; in value.{assert_to_dot[1]}"
            else:
                assert_to = f"'{assert_to}'" if isinstance(assert_to, str) else assert_to
            returned = test['status']['reason']
            returned = f"'{returned}'" if isinstance(returned, str) else returned
            
            rich.print(f"[bold yellow]{' '*12}- assertion: {test['assertion']}[/bold yellow]")
            rich.print(f"[bold yellow]{' '*12}- assert to: {assert_to}[/bold yellow]")
            rich.print(f"[bold yellow]{' '*12}- returned: {returned}[/bold yellow]")
            if test['status']['passed']:
                rich.print(f"[bold green]{' '*12}-Test: passed[/bold green]")
            else:
                rich.print(f"[bold red]{' '*12}-Test: failed[/bold red]")

def capybara_output(args):
    rich.print(f"[bold blue]Litio[/bold blue][bold green]@[/bold green][bold blue]{info.__version__}[/bold blue] - [bold green]{args.title}[/bold green]")
    for group in args.groups:
        group_name = group["name"]
        group_name = group_name.replace("-", " ").replace("_", " ")
        group_name = group_name.title()
        passed_tests = len([test for test in group["tests"] if test.get('status', {'passed': False}).get('passed')])
        not_passed_tests = len([test for test in group["tests"] if not test.get('status', {'passed': True}).get('passed')])
        ignored = len([test for test in group["tests"] if test.get("ignore")])
        print(f"[{group_name}] ✔️  Passed {passed_tests} ❌ Errors {not_passed_tests}  ⚠️  Ignored {ignored}")
        most_largest_name = 0
        for _group in args.groups:
            for test in _group["tests"]:
                if len(test['name']) > most_largest_name:
                    most_largest_name = len(test['name'])
        for test in group["tests"]:
            if test.get('ignore'):
                tabs = most_largest_name + 6 - len(test['name'])
                tabs = ' '*tabs
                print(f"{' '*3}- {test['name']}{tabs}⚠️  Ignored")
                continue
            if test['status']['passed']:
                tabs = most_largest_name + 6 - len(test['name'])
                tabs = ' '*tabs

                print(f"{' '*3}- {test['name']}{tabs}{'✔️  Passed'}")
            else:
                tabs = most_largest_name + 6 - len(test['name'])
                tabs = ' '*tabs
                print(f"{' '*3}- {test['name']}{tabs}{'❌  Error'} \"{test['status']['reason']}\"")
        print() # new line

def output(args, style, outputs={}):
    outputs.get(style, capybara_output)(args)


import os, argparse, importlib, subprocess
from git import Repo
import json
from . import litio, info, output, utils, tester, ai
import tarfile


def uncompress(source, dest):
    with tarfile.open(source, "r:gz") as tar:
        tar.extractall(path=dest)

def purge_dir(dir):
    if not os.path.exists(dir):
        return
    for f in os.listdir(dir):
        if os.path.isdir(os.path.join(dir, f)):
            purge_dir(os.path.join(dir, f))
            os.rmdir(os.path.join(dir, f))
        else:
            os.remove(os.path.join(dir, f))

def prepare_for_install():
    if not os.path.exists(f'./litio'):
        os.mkdir(f'./litio')
    if not os.path.exists(f'./litio/modules'):
        os.mkdir(f'./litio/modules')
    if not os.path.exists(f'./litio/modules/.dev'):
        os.mkdir(f'./litio/modules/.dev')

def install_module(args, get_module):
    module = args.module
    version = None
    if module.endswith('.tar.gz'):
        prepare_for_install()
        name = ".".join(module.split(".")[:-2])
        name = name.split("/")[-1]
        no_version_name = "-".join(name.split("-")[:-1])
        print("Installing module from tarball")
        for f in os.listdir('./litio/modules/.dev'):
            if no_version_name in f:
                purge_dir(f'./litio/modules/.dev/{f}')
                os.rmdir(f'./litio/modules/.dev/{f}')
        uncompress(module, f'./litio/modules/.dev/{name}')
        if 'requirements.txt' in os.listdir(f'./litio/modules/.dev/{name}/'):
            print(f"Installing requirements for module {name}...")
            subprocess.run(['pip', 'install', '-r', f'./litio/modules/.dev/{name}/requirements.txt'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        print(f"Module {name} installed successfully")
        return
        
    if '@' in module:
        module, version = module.split('@')
    prepare_for_install()
    if args.upgrade and os.path.exists(f'./litio/modules/{module}'):
        purge_dir(f'./litio/modules/{module}')
        os.rmdir(f'./litio/modules/{module}')
    if os.path.exists(f'./litio/modules/{module}'):
        print(f"Module {module} already installed")
        return
    if not os.path.exists(f'./litio/modules/{module.split("/")[0]}'):
        os.mkdir(f'./litio/modules/{module.split("/")[0]}')
    os.mkdir(f'./litio/modules/{module}')
    
    
    repository = f'https://github.com/{module}.git'
    print(f"Installing module {module}...")
    repo = Repo.clone_from(repository, f'./litio/modules/{module}')
    if args.branch:
        try:
            repo.git.checkout(args.branch)
        except:
            print(f"Branch '{args.branch}' not found")
            print(f'Installing module at default branch...')
    if version:
        try:
            repo.git.checkout(version)
        except:
            print(f"Version '{version}' not found")
            print(f'Installing module at last version...')
    if 'requirements.txt' in os.listdir(f'./litio/modules/{module}'):
        print(f"Installing requirements for module {module}...")
        subprocess.run(['pip', 'install', '-r', f'./litio/modules/{module}/requirements.txt'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    print(f"Module {module} installed successfully")

def remove_module(args, get_module):
    if not os.path.exists(f'./litio/modules/{args.module}'):
        print(f"Module {args.module} not installed")
        return
    
    purge_dir(f'./litio/modules/{args.module}')
    os.rmdir(f'./litio/modules/{args.module}')
    print(f"Module {args.module} removed successfully")

def pack_module(args, get_module):
    required_files = ["litio.py", "litio.json"]
    for file in required_files:
        if not os.path.exists(file):
            print(f"Required file {file} not found")
            return
    ignore = ["dist"]
    if os.path.exists('./.litioignore'):
        ignore += open('./.litioignore', 'r').read().splitlines()
    package_data = json.load(open('./litio.json', 'r'))
    if not package_data.get('name'):
        print('Missing name in litio.json')
        return
    if not package_data.get('version'):
        print('Missing version in litio.json')
        return
    if not os.path.exists('./dist/'):
        os.mkdir('./dist/')
    else:
        purge_dir('./dist/')
    tar = tarfile.open(f'./dist/{package_data["name"]}-{package_data["version"]}.tar.gz', 'w:gz')
    for file in os.listdir('./'):
        if file in ignore and not file in required_files:
            continue
        tar.add(file)
    tar.close()
    print("Module packed successfully")

options = {
    'version': {
        'option': '--version',
        'aliases': ['-v'],
        'action': 'version',
        'version':'%(prog)s v{}'.format(info.__version__)
    }
}

sub_commands = {
    'pack': {
        'command': 'pack',
        'help': 'pack a litio module',
        'controller': pack_module,
        'options': {
            
        }
    },
    'run': {
        'command': 'run',
        'help': 'run a litio test',
        'controller': litio.run,
        'options': {
            'verbose': {
                'option': '--verbose',
                'aliases': ['-V'],
                'dest': 'verbose',
                'help': 'enable verbosity',
                'required': False,
                'action': argparse.BooleanOptionalAction
            },
            'ai': {
                'option': '--ai',
                'dest': 'ai',
                'help': 'enable/disable ai',
                'required': False,
                'action': argparse.BooleanOptionalAction,
                'default': True,
            },
            'output': {
                'option': '--output',
                'aliases': ['-o'],
                'dest': 'output',
                'help': 'output style',
                'required': False,
                'default': None
            }
        }
    },
    'install': {
        'command': 'install',
        'help': 'install a litio module',
        'options': {
            'module': {
                'option': 'module',
                'help': 'module to install'
            },
            'upgrade': {
                'option': '--upgrade',
                'aliases': ['-u'],
                'help': 'upgrade a litio module',
                'required': False,
                'action': 'store_true'
            },
            'branch': {
                'option': '--branch',
                'aliases': ['-b'],
                'dest': 'branch',
                'default': None,
                'help': 'branch to get the module from',
                'required': False                
            }
        },
        'controller': install_module
    },
    'uninstall': {
        'command': 'uninstall',
        'help': 'uninstall a litio module',
        'options': {
            'module': {
                'option': 'module',
                'help': 'module to uninstall'
            }
        },
        'controller': remove_module
    }

}

modules_info = {
    'output': {
        'classic': output.output_classic,
        'capybara': output.capybara_output,
        'output': output.output
    },
    'options': options,
    'sub_commands': sub_commands,
    'utils': {
        'extract_function_code': utils.extract_function_code,
        'Args': utils.Args,
        'OutputArgs': utils.OutputArgs,
        'params_to_dict': utils.params_to_dict,
        'eval_params_values': utils.eval_params_values,
        'Kwargs': utils.Kwargs  
    },
    'tester': {
        'test': tester.test
    },
    'ai': {
        'BugFixer': ai.BugFixer,
        'fix_bug': ai.fix_bug
        
    }
}
def load_modules(safe_mode):
    if not os.path.exists('./litio/modules') or safe_mode:
        return modules_info
    modules = os.listdir(f'./litio/modules')
    modules = [module for module in modules if module != ".dev"]
    for author in modules:
        for module in os.listdir(f'./litio/modules/{author}'):
            if not os.path.exists(f'./litio/modules/{author}/{module}/litio.py'):
                continue
            spec = importlib.util.spec_from_file_location(f'{author}-{module}', f'./litio/modules/{author}/{module}/litio.py')
            modulelib = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(modulelib)
            for key, value in modulelib.litio.items():
                modules_info[key].update(value)
    for module in os.listdir('./litio/modules/.dev'):
        if not os.path.exists(f'./litio/modules/.dev/{module}/litio.py'):
            continue
        spec = importlib.util.spec_from_file_location(f'.dev-{module}', f'./litio/modules/.dev/{module}/litio.py')
        modulelib = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(modulelib)
        for key, value in modulelib.litio.items():
            modules_info[key].update(value)
    return modules_info


def addOneWR(n):
    return n + 1

class Person:
    def __init__(self, name, age):
        self.__name = name
        self.__age = age
    def get_name(self):
        return self.__name
    def get_age(self):
        return self.__age


import rich, yaml, os

def run(args, get_module):
    if not os.path.exists('litio.yml'):
        print("❌ No config file 'litio.yml' found")
        exit(1)
    data = open('./litio.yml','r').read()
    data = yaml.safe_load(data)
    tests_output_args = get_module('utils').OutputArgs(title=data["name"])
    for test, test_data in data["tests"].items():
        path = test_data['path']
        tests_output_args.add_group({
            "name": test,
            "path": path,
            "tests": []
        })
        for function in test_data['functions']:
            function_name = list(function.keys())[0]
            test_dict = {
                'name': function_name,
                'verbose': args.verbose or function[function_name].get('verbose')
            }
            if function[function_name].get('ignore'):
                test_dict['ignore'] = True
                tests_output_args.add_test(test_dict, test)
                continue
            
            argsDitctionary = {
                "file": path,
                "function": function_name
            }
            if function[function_name].get('inputs'):
                inputs = eval(str(function[function_name]["inputs"]))
            else:
                inputs = {}
            argsDitctionary.update({"params":inputs})
            if function[function_name].get('expected'):
                if function[function_name]["expected"].get('self'):
                    ... # Coming soon, acces to class self attributes
                if not function[function_name]["expected"].get('value'):
                    keys = list(function[function_name]["expected"].keys())
                    keys.pop(keys.index('comparator'))
                    value_name = keys[0]
                    assert_to = function[function_name]["expected"][value_name]
                    function[function_name]["expected"]["value"] = function[function_name]["expected"][value_name]
                    indexes = value_name.split('.')[1:]
                    argsDitctionary.update({
                        "assertion":function[function_name]["expected"]["comparator"],
                        "assert_to_dot": [assert_to,indexes]
                        })
                 
                else:
                    argsDitctionary.update({"assertion":function[function_name]["expected"]["comparator"], "assert_to":function[function_name]["expected"]["value"]})
            if function[function_name].get('instance'):
                argsDitctionary.update({"instance_params":function[function_name]["instance"]})
            argsDitctionary.update({"print_return":function[function_name].get('print-return') or args.verbose or function[function_name].get('verbose')})
            
            test_dict.update(argsDitctionary)
            new_params = {}
            for key, value in test_dict['params'].items():
                if type(value) == dict:
                    new_params[key] = value.copy()
                else:
                    new_params[key] = value
            test_dict['params'] = new_params
            argsToMain = get_module('utils').Args(argsDitctionary)

            main_response = get_module('tester').test(argsToMain, get_module)
            if not main_response:
                test_dict["status"] = {"passed": False, 'reason': "Function not found"}
                tests_output_args.add_test(test_dict, test)
                continue
            if isinstance(main_response, str) and args.ai:
                fixed_bug = get_module('ai').fix_bug(function_name=function_name, path=path, data=data, inputs=inputs, main_response=main_response, function=function, args=args)
                if not fixed_bug:
                    test_dict["status"] = {"passed": False, 'reason': "Could not fix bug"}
                    tests_output_args.add_test(test_dict, test)
                    continue
                if isinstance(fixed_bug, str):
                    test_dict['status'] = {"passed": False, 'reason': fixed_bug}
                    tests_output_args.add_test(test_dict, test)
                    continue
                to_print = main_response[0] if not test_dict.get('assert_to_dot') else main_response[2]
                assertion = fixed_bug[1]
            elif not main_response[1] and args.ai:
                fixed_bug = get_module('ai').fix_bug(function_name=function_name, path=path, data=data, inputs=inputs, main_response=main_response, function=function, args=args)
                if not fixed_bug:
                    test_dict["status"] = {"passed": False, 'reason': "Could not fix bug"}
                    tests_output_args.add_test(test_dict, test)
                    continue
                if isinstance(fixed_bug, str):
                    test_dict['status'] = {"passed": False, 'reason': fixed_bug}
                    tests_output_args.add_test(test_dict, test)
                    continue
                to_print = main_response[0] if not test_dict.get('assert_to_dot') else main_response[2]
                assertion = fixed_bug[1]
            else:
                to_print = main_response[0] if not test_dict.get('assert_to_dot') else main_response[2]
                assertion = main_response[1]
            
            if not assertion:
                comparator = function[function_name]['expected']['comparator']
                test_dict['status'] = {"passed": False, 'reason': f"{to_print} not {comparator} {function[function_name]['expected']['value']}"}
                tests_output_args.add_test(test_dict, test)
                continue
            test_dict['status'] = {"passed": assertion, 'reason': to_print}
            tests_output_args.add_test(test_dict, test)
    
    styles = get_module('output', True).copy()
    styles.pop('output')
    get_module('output').output(tests_output_args, args.output if args.output else data.get("output-style", "capybara"), styles)



def addOne(n):
    return n + 1
def addOneWR(n):
    return n + 1
def concatenate_lists(a, b):
    return a + b
def concatenate_dicts(a,b):
    a.update(b)
    return a
def pow(base, exponent):
    return base**exponent
  
class Hello:
    def __init__(self,name):
        self.name = name
    def sum_lists(self, a, b):
        a.extend(b)
        return a
    @classmethod
    def print_hello(self,name):
        prt = f"hello {name}"
        return prt

class World:
    def __init__(self,name):
        self.name = Hello(name)
        self.__world_id = 1243
    def hello(self):
        return Hello.print_hello(self.name)
    
def generate_world(name):
    return World(name)


import importlib.util

def test(args, get_module):
        module_name = 'module'
        spec = importlib.util.spec_from_file_location(module_name, args.file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        if "." in args.function:
            module.__dict__
            _class = module.__dict__.get(args.function.split(".")[0], None)
            if not _class:
                return None
            function = _class.__dict__.get(args.function.split(".")[1])
            if not function:
                return None
            if type(function) == classmethod:
                function_type = "classmethod"
            else:
                function_type = "method"
            function_params = function.__annotations__
        else:
            function_type = "function"
            function_params = module.__dict__.get(args.function, None)
            if not function_params:
                return None
            function_params = function_params.__annotations__
        if type(args.params) == list:    
            params = get_module('utils').params_to_dict(args.params)
            params = get_module('utils').eval_params_values(params,function_params)
        else:
            params = args.params
        to_return_assert = None
        use_dot_to_return = False
        if function_type == "method":
            init_params = module.__dict__.get(args.function.split('.')[0]).__init__.__annotations__
            if type(args.instance_params) == list:    
                instance_params = get_module('utils').params_to_dict(args.instance_params)
                instance_params = get_module('utils').eval_params_values(instance_params,init_params)
            else:
                instance_params = args.instance_params
        try:
            if function_type == "function":
                fun = getattr(module,args.function)
                try:
                    to_return = fun(**params)
                except Exception as e:
                    return str(e)

            elif function_type == "method":
                class_name = args.function.split(".")[0]
                method_name = ".".join(args.function.split(".")[1:])
                _class = getattr(module,class_name)
                instance = _class(**instance_params)
                try:
                    to_return = getattr(instance,method_name)(**params)
                except Exception as e:
                    return str(e)

            elif function_type == "classmethod":
                _class = getattr(module,args.function.split(".")[0])
                fun = getattr(_class,args.function.split(".")[1])
                try:
                    to_return = fun(**params)
                except Exception as e:
                    return str(e)
            
            if args.__dict__.get('assert_to'):
                assert_to = args.__dict__.get('assert_to')
            else:
                assert_to, indexes = args.assert_to_dot
                use_dot_to_return = True
                dot_to_return = to_return
                last_return = to_return
                for index in indexes:
                    if isinstance(last_return, (list, tuple, set)):
                        if index.count(';') == 1:
                            index_0 = int(index.split(';')[0]) if index.split(';')[0].replace('-','').isdigit() else None
                            index_1 = int(index.split(';')[1]) if index.split(';')[1].replace('-','').isdigit() else None
                            last_return = last_return[index_0:index_1]
                        elif index.count(';') == 2:
                            index_0 = int(index.split(';')[0]) if index.split(';')[0].replace('-','').isdigit() else None
                            index_1 = int(index.split(';')[1]) if index.split(';')[1].replace('-','').isdigit() else None
                            index_2 = int(index.split(';')[2]) if index.split(';')[2].replace('-','').isdigit() else None
                            last_return = last_return[index_0:index_1:index_2]
                        else:
                            last_return = last_return[int(index)]
                    elif type(last_return) == dict:
                        last_return = last_return[index]
                    else:
                        try:
                            last_return = last_return.__dict__[index]
                        except KeyError:
                            return [f'{last_return.__class__.__name__} has no attribute {index}']
                to_return = last_return
            if args.assertion != None:
                if assert_to == None:
                    to_return_assert = ('No assertion to perform')
                if args.assertion == "Equals":
                    to_return_assert = (to_return == assert_to)
                elif args.assertion == "NotEquals":
                    to_return_assert = (to_return != assert_to)
                elif args.assertion == "Greater":
                    to_return_assert = (to_return > float(assert_to))
                elif args.assertion == "GreaterOrEquals":
                    to_return_assert = (to_return >= float(assert_to))
                elif args.assertion == "Less":
                    to_return_assert = (to_return < float(assert_to))
                elif args.assertion == "LessOrEquals":
                    to_return_assert = (to_return <= float(assert_to))
                elif args.assertion == "In":
                    to_return_assert = (to_return in eval(assert_to))
                elif args.assertion == "NotIn":
                    to_return_assert = (to_return not in eval(assert_to))
                elif args.assertion == "Is":
                    to_return_assert = (to_return is eval(assert_to))
                elif args.assertion == "IsNot":
                    to_return_assert = (to_return is not eval(assert_to))
                elif args.assertion == "IsNone":
                    to_return_assert = (to_return is None)
                elif args.assertion == "IsNotNone":
                    to_return_assert = (to_return is not None)
                elif args.assertion == "IsInstance":
                    to_return_assert = (to_return.__class__ == getattr(module,assert_to))
                elif args.assertion == "IsNotInstance":
                    to_return_assert = (to_return.__class__ != getattr(module,assert_to))
                else:
                    to_return_assert = False
                    to_return = (f"Assertion '{args.assertion}' not found")
            return [dot_to_return, to_return_assert, to_return] if use_dot_to_return else [to_return,to_return_assert]
        except Exception as e:
            return [dot_to_return, False, to_return] if use_dot_to_return else [to_return, False]
            


import argparse, yaml, os
try:     
    from utils import modules, info
except ModuleNotFoundError:
    from .utils import modules, info


def litio():
    parser = argparse.ArgumentParser(description='A command line function tester')
    data = {
        'safe-mode': False
    }
    if os.path.exists('./litio.yml'):
        data = open('./litio.yml', "r").read()
        data = yaml.safe_load(data)
    
    modules_info = modules.load_modules(data.get('safe-mode'))
    def get_module(module_name, info=False):
        if info:
            return modules_info[module_name]
        return modules_info['utils']['Kwargs'](**modules_info[module_name])
    
    for option in list(modules_info['options'].values()): # Load options and add them to the parser
        name = option['option']
        aliases = option.get('aliases', [])
        flags_or_name = [name] + aliases
        if option.get('aliases'):
            option.pop('aliases')
        option.pop('option')
        parser.add_argument(*flags_or_name, **option)
    subparsers = parser.add_subparsers(help='sub-command help')
    for sub_command in list(modules_info['sub_commands'].values()):
        command = sub_command['command']
        sub_parser = subparsers.add_parser(name=command, help=sub_command.get('help'))
        for option in list(sub_command['options'].values()): # Load options and add them to the sub parser
            name = option['option']
            aliases = option.get('aliases', [])
            flags_or_name = [name] + aliases
            if option.get('aliases'):
                option.pop('aliases')
            option.pop('option')
            sub_parser.add_argument(*flags_or_name, **option)
        sub_parser.set_defaults(func=sub_command['controller'])
    
    args = parser.parse_args()
    if args.__dict__.get('func') is None:
        parser.print_help()
        exit(0)    
    args.func(args, get_module)
    exit(0)

if __name__ == '__main__':
    litio()



from setuptools import setup,find_packages

with open('README.md', 'r') as reader:
    readme = reader.read()

requirements = [
    "PyYAML==6.0.1",
    "rich==13.7.0",
    "openai==1.3.5",
    "python-dotenv==1.0.0",
    "gitpython==3.1.40"
]

setup(
    author='Lizardwine',
    author_email='lizardwine@hotmail.com',
    name='litio',
    description='A simple tool for testing',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/lizardwine/litio',
    project_urls={
        "Bug Tracker": "https://github.com/lizardwine/litio/issues",
    },
    license='GPL v3.0',
    version='1.6.1',
    keywords=['testing', 'tester'],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development'
    ],
    entry_points={
        'console_scripts': [
            'litio = litio.__main__:main'
        ]
    },
    package_dir={"": "src"},
    install_requires=requirements,
    packages=find_packages(where="src"),
    python_requires=">=3.6",
)